//---------------------------------------------------------------------
//
// "techorate_echinodermata.pov" -- Torsten Crass, 2008
//
// A demo application of "techorator.inc", a POV-Ray macro package 
// for applying arbitrary itmes to arbitrary objects' surfaces in 
// order to make them look "technical"
//
// This scene file, however, demonstrates that the techorator macro
// can also be employed for creating organically looking things like,
// in this case, sea urchins and sea cucumbers (both of which, like
// starfish, belonging to the phylum "Echinodermata").
//
// In particular, this scene demonstrates the use of completely
// random placements of techoration items as well as the use of
// custom re-definitions of the getTechorationItem macro.
//
//---------------------------------------------------------------------



// For demonstrating the techorator, we have to include it, ne c'est pas?
#include "techorator.inc"

// We'll also make use of some pre-defined transformation macros
#include "transforms.inc"



//--- Constant definitions --------------------------------------------

// Random seed to be used in this scene (has nothing to do with
// random streams used internally by the techorator!)
#declare rs = seed(0);

// Sea urchin's number of stings
#declare nStings = 5000;
// Sea cucumber's number of chaetae
#declare nChaeta = 500;



//--- Texture definitions ---------------------------------------------

// Sandy seafloor
#declare T_Sand = texture {
  pigment {color rgb <1, 0.9, 0.5>}
  finish {
    ambient 0.4
    diffuse 0.8
    brilliance 1
  }
  normal {
    ripples 2
    turbulence 0.5
    scale 3
    translate <-20, 0, -20>
  }
  
}

// A greenish texture for the sea urchin's shell
#declare T_SeaUrchinShell = texture {
  pigment {color rgb <0.5, 0.8, 0.5>}
  normal {
    radial 0.5
    frequency 90
  }
}

// The sea urching gets nice purple stings
#declare T_Sting = texture {
  pigment {color rgb <0.5, 0, 0.5>}
  finish {
    phong 1
    phong_size 5
  }
}

// The sea cucumber's body texture
#declare T_SeaCucumber = texture {
  pigment {color rgb <0.5, 0.25, 0.0>}
}

// Texture for the sea cucumber's hair
#declare T_Chaeta = texture {
  pigment {color rgb <1.0, 0.75, 0.2>}
  finish {
    phong 1
    phong_size 5
  }
}



//--- Basic scene setup -----------------------------------------------

// Main light
light_source {
  <400, 100, -400> 
  color rgb 1.0
}

// For slightly illuminate shadowd regions 
light_source {
  <-400, 0, -100>
  color rgb 0.3
  shadowless
}

// Camera
camera {
  location  <-3, 5, -10>
  look_at   <0, 2, 0>
  right     x*image_width/image_height
}

// Blue-ish underwater haze
fog {
  distance 40
  color rgb<0.0, 0.5, 0.6>
}

// The sea floor
plane {y 0 texture {T_Sand} }



//--- Sea urchin ------------------------------------------------------

// A single, cone-shaped sting
#declare sting = cone {-0.1*y, 0.05, y, 0
  texture {T_Sting}
}

// The sting is the only type of item we wanna techorate with
#declare items = array[1] {
  sting
}
// Hence this should yield 1
#declare nItems = dimension_size(items, 1);

// Do not scale the stings
#declare scaling = array[nItems] {
  <0, 0, 0, 0>,
}
 
// Allow each sting to be randomly rotated from 0 to 360 degrees
// in one-degree steps (360/360)
#declare modifications = array[nItems] {
  <0, 360, 360, 0, 0>,
}

// Pass items and what to do with them to the techorator
setTechorationItems(items, scaling, modifications)
  
// Tell that we want all items to be used with equal probabilities
// (Well, since there's only on item, any greater-than-zero number will do)
#declare weights = array[nItems] {
  1
}
setTechorationWeights(weights)

// Set up a full spherical grid. The indidvidual grid nodes, however, will not
// be further considered since we're going to use full random techoration.
setTechorationDimensions(TECHORATE_SPHERICAL_GEOMETRY, 0, 360, 0, -90, 90, 0, 20, 15, 20)

// Re-definition of the techorator macro that yields a new item. Normaly directly 
// returns the result of getTechorationItem_techorator_native(ttype), but we wann add
// some rotation to randomly divert the stings from the surface normals
// Parameters:
//    ttype:    index of desired item type in item array
//    i, j:     grid position (in case you wanna return node-specific elements)
//    sVec:     vector pointing at the item's final position on the techorated 
//              object's surface
//    nVec:     surface normal at the item's final position
#macro getTechorationItem(ttype, i, j, sVec, nVec)
  // Just get the item...
  #local tObj = getTechorationItem_techorator_native(ttype);
  // ...and rotate it a little bit around x axis
  object {tObj rotate 30*rand(rs)*x }
#end


// Declaration of the sea urchin's basic shape
#declare seaUrchinShape = sphere {0, 1 scale <2, 1.5, 2>};

#declare seaUrchin = union {
  // The "naked" sea urchin shell
  object {seaUrchinShape texture {T_SeaUrchinShell}}
  // Now apply randomly techorate with stings
  techorateRandomly(seaUrchinShape, nStings)
}

// Move the sea urchin to its final position
object {seaUrchin translate <4, 1.5, 0>}


//--- Sea cucumber ----------------------------------------------------


// Another re-definition of the techorator macro that yields 
// individual items for techoration; this time: returns a
// sphere sweep the on end of which will point parallel to the
// normal vector, while the other end's direction will just follow
// gravity (negative z direction). However, since we gonna turn the
// sea cucumber upright for techoration, "down" becomes "left" 
// (negative x) when the macro gets called.
#macro getTechorationItem(ttype, i, j, sVec, nVec)
  // Local constants
  #local d = 1.5; // influences diameter size of chaetae
  #local s = 0.8; // final scaling factor

  #local tObj = sphere_sweep {cubic_spline 6,
    -0.2*nVec, d*0.1,       // -
    0, d*0.1,               //  | these three points form a line parallel to normal
    0.4*nVec, d*0.08        // -
    0.7*nVec - 1*x, d*0.05, // -
    0.4*nVec - 2*x d*0.02,  //  | here the influence of gravity increases
    0.0*nVec - 2.5*x d*0.00 // -
    texture {T_Chaeta}      // Of course, a chata should look like a chaeta
  };

  // Now we must "reverse" the transformations which will be performed
  // by the techorator on the chaeta since it assumes the item to
  // point in positive y direction, not parallel to the normal vector
  object {tObj
    // Adjust size
    scale s 
    // First let "normal" point towards negative z...
    Reorient_Trans(<nVec.x, 0, nVec.z>, -z)
    // ...then make it look into positive y direction
    Reorient_Trans(<0, nVec.y, -1>, y)
  }
#end

// Allow each sting to be randomly rotated from 0 to 360 degrees
// in one-degree steps (360/360)
#declare modifications = array[nItems] {
  <0, 0, 0, 0, 0>,
}

// Pass items and what to do with them to the techorator
setTechorationItems(items, scaling, modifications)


// The sea cucumber's body is made from a (vertical) cylinder with
// (visible: semi-) spheres at both ends
#declare seaCucumberShape = union {
  cylinder {-4*y, 4*y, 1}
  sphere {-4*y 1}
  sphere {4*y 1}
}

#declare seaCucumber = union {
  // The sea cucumber's naked body
  object {seaCucumberShape texture {T_SeaCucumber}}

  // Create a full cylindrical grid for the main part of the body
  setTechorationDimensions(TECHORATE_CYLINDRICAL_GEOMETRY, 0, 360, 0, -4, 4, 0, 20, 0, 20)
  // Randomly place a lot of chaetae
  techorateRandomly(seaCucumberShape, nChaeta)

  // Now create a partial spherical grid for techorating the lower semi-sphere
  setTechorationDimensions(TECHORATE_SPHERICAL_GEOMETRY, 0, 360, 0, -90, 0, 0, 20, 20, 20)
  union {
    // Move the sea cucumber's shape to get the lower semi-spheres center
    // into origin, and techorate it
    techorateRandomly(object{seaCucumberShape translate 4*y}, nChaeta/4)
    // Now move the union of all chaetae back to the body shape's
    // lower end
    translate -4*y
  }

  // The same for the upper semi-sphere
  setTechorationDimensions(TECHORATE_SPHERICAL_GEOMETRY, 0, 360, 0, 0, 90, 0, 20, 20, 20)
  union {
    techorateRandomly(object{seaCucumberShape translate -4*y}, nChaeta/4)
    translate 4*y
  }
}

// Transform the sea cucumber into its final position
object {seaCucumber scale 0.9 translate 4*y rotate -0*x rotate 90*z rotate 60*y translate <-2.5, 1.5, 0>}
