//---------------------------------------------------------------------
//
// "techorate_starship.pov" -- Torsten Crass, 2008
//
// A demo application of "techorator.inc", a POV-Ray macro package
// for applying arbitrary itmes to arbitrary objects' surfaces in
// order to make them look "technical"
//
// This scene file demonstrates how to use the techorator for
// creating objects looking similar to starships as seen
// in many science-fiction movies
//
//---------------------------------------------------------------------



// For demonstrating the techorator, we have to include it, ne c'est pas?
#include "techorator.inc"

// If you just wanna have a glance at the items used for techorating
// the starship, set this to 1:
#declare showItems = 0;

// If povray has not been called by povmov.pl, we manually specify
// which (relative) time index we wanna render
#ifndef(relMovieTime)
  #declare relMovieTime = 0.75;
#end


//--- Texture definitions ---------------------------------------------

// The vessel's primary hull texture
#declare T_Hull = texture {
  pigment {color rgb <1, 0.9, 0.55>}
  finish {
    ambient 0.1
    diffuse 0.8
    brilliance 0.5
    phong 0.2
    phong_size 50
    metallic
  }
}

// A fancy red hull texture
#declare T_RedHull = texture {
  pigment {color rgb <0.4, 0.1, 0.0>}
  finish {
    ambient 0.1
    diffuse 0.8
    brilliance 1
    phong 0.2
    phong_size 1
    metallic
  }
}

// Texture for light windows
#declare T_Light = texture {
  pigment {color rgb <1, 1, 0.9>}
  finish {
    ambient 1
  }
}

// The planet's surface
#declare T_Planet = texture {
  pigment {
    marble
    color_map {
      [0.3 color rgb <0, 1, 0.5>]
      [0.5 color rgb <0, 0.5, 0>]
    }
    turbulence 1
    scale 2000
  }
  finish {
    ambient 0.1
    brilliance 0.2
  }
}

// An auto-scaling white-stars-on-black-background texture
#declare P_Space = pigment {
  bozo
  color_map {
    [0.8 color rgb 0]
    [0.85 color rgb 1]
  }
  scale 1.5/max(image_width, image_height)
}


//--- Basic scene setup -----------------------------------------------

// Main illumination
light_source {
  <400000, 100000, -400000>
  color rgb 1.0
}

// For slightly illuminate shadowd regions
light_source {
  <-400, -100, -100>
  color rgb 0.1
  shadowless
}

// The final frontier...
sky_sphere {
  pigment {P_Space}
}


//--- Macro definitions -----------------------------------------------

// Creates a thin, disk-like cylinder with a glowing texture of color col;
// glow intensity is i from center to r1 and linearly decreases from i at
// r1 down to 0 at r2. The cylinder's axis reaches from the origin to
// <0, -0.01, 0>.
#macro createGlow(r1, r2, col, i)
  disc {<0,0,0>, -0.01*z, r2
    texture {
      pigment {
        onion
        color_map {
          [r1/r2 color rgbt <col.red, col.green, col.blue, 1-i>]
          [1     color rgbt <col.red, col.green, col.blue, 1>]
        }
      }
      finish {
        ambient 1
        diffuse 0
      }
      scale r2
    }
    no_shadow
  }
#end



//--- Some object definitions -----------------------------------------

// A spline defining how the camera will move
// during the short demo video created by povmov.pl
#declare camera_path = spline {cubic_spline
  -5, <25, -22, 50>
  0, <20, -10, 30>
  5, <15, 2, 10>
  7, <12, 2, 0>
  9, <8, 1, -5>
  11, <0, 0, -8>
  13, <-8, 0, -8>
  15, <-12, 1, -6>
  20 <-30, 10, -6>
  25 <-48, 18, -6>
}


//--- The vessel ---

// The vessel's basic shape
#declare hull = blob {
  threshold 0.4
  sphere {0, 10, 1 scale <1, 0.2, 0.5>}
  sphere {0, 6, 1 scale <0.5, 1.2, 0.2> translate <-4, -1, 0>}
  sphere {0, 1.25, 1 scale <1, 0.75, 0.75> translate <4, 0.75, 1>}
  sphere {0, 5, -1 scale <1, 0.5, 1> translate <2, -2, 0>}
  texture {T_Hull}
  scale 2
}

// An object used for cutting off a piece of the vessel's rear part
// to make room for the engines
#declare engine_cutoff1 = prism {linear_spline -2, 2, 4
  <0, 0> <-10, 2> <-10, -10> <5, -10>
  texture {T_Hull}
}

// Another object used for carving out the engine location
#declare engine_cutoff2 = intersection {
  object {hull}
  prism {linear_spline -2, 2, 4
    <0, 0> <-10, 2> <-10, -10> <5, -10>
    translate <-8, 0, -2.25>
  }
  texture {T_Hull}
  scale 0.99
}

//
#declare vessel = difference {
  object {hull}
  union {
    object {engine_cutoff1 translate <-9, 0, -2.25>}
    object {engine_cutoff1 scale <1, 1, -1> translate <-9, 0, 2.25>}
    object {engine_cutoff2 translate <0, 0, 0>}
    object {engine_cutoff2 scale <1, 1, -1> translate <0, 0, 0>}
  }
  translate -0.5*y
  rotate 90*z
}



//--- The engines

// The nozzle
#declare nozzle = union {
  lathe { quadratic_spline 8,
    <2, -2> <2, 4> <0.5, 4.5> <0.5, 5> <1, 6> <2.5, 9> <2.4, 8> <0, 6>
  }
  cylinder {-2*y, 4*y, 2}
  scale 0.3
  rotate 90*z
  texture {T_Hull}
}

// The exhaust flame
#declare flame = lathe { cubic_spline 7
  <2, 5> <2.5, 9> <2.8, 10> <2, 12> <1, 15> <0, 18> <-1, 20>
  texture {
    pigment {
      gradient y
      color_map {
        [0 color rgbt <0, 1, 1, 0.3>]
        [0.95 color rgbt <0, 1, 1, 1>]
      }
      scale 9
      translate 9*x
    }
    finish {
      ambient 0.8
    }
  }
  no_shadow  rotate 90*z
  scale 0.3
  
}



//--- Items for techorating the starship ------------------------------

#declare plate1 = union {
  union {
    cylinder {<-0.4, 0, 0> <-0.4, 0, 0.4> 0.05}
    cylinder {<-0.3, 0, 0> <-0.3, 0, 0.4> 0.05}
    cylinder {<-0.2, 0, 0> <-0.2, 0, 0.4> 0.05}
    cylinder {<-0.1, 0, 0> <-0.1, 0, 0.4> 0.05}
    texture {T_Hull}
  }
  intersection {
    box {<-0.5, -1, -0.1> <0.5, 1, -0.4> texture {T_Hull}}
    cylinder {-0.5*z, 0, 3 translate -2.95*y texture {T_RedHull}}
  }
}

#declare plate2 = union {
  difference {
    box {<-0.4, -0.1, -0.4> <0.4, 0.05, 0.4>}
    union {
      box {<-0.3, -0.2, -0.3> <0.3, 0.1, 0>}
      plane {-x-z, -0.3}
    }
  }
  sphere {<-0.0, 0, 0.2> 0.1}
  sphere {<-0.2, 0, 0.2> 0.1}
  texture {T_Hull}
}

#declare plate3 = union {
  union {
    cylinder {<-0.4, 0, 0> <-0.4, 0, 0.4> 0.05}
    cylinder {<-0.3, 0, 0> <-0.3, 0, 0.4> 0.05}
    cylinder {<-0.2, 0, 0> <-0.2, 0, 0.4> 0.05}
    cylinder {<-0.1, 0, 0> <-0.1, 0, 0.4> 0.05}
    texture {T_Hull}
  }
  intersection {
    box {<-0.5, -1, -0.1> <0.5, 1, -0.4> texture {T_Hull}}
    cylinder {-0.5*z, 0, 3 translate -2.95*y texture {T_Hull}}
  }
}

#declare windows = union {
  box {<-0.4, -0.05, 0.2> <-0.2, 0.5, 0.3>}
  box {<0.2, -0.05, 0.2> <0.4, 0.5, 0.3>}
  box {<-0.4, -0.05, -0.2> <-0.2, 0.5, -0.3>}
  box {<-0.1, -0.05, -0.2> <0.1, 0.5, -0.3>}
  rotate 90*y
  texture {T_Light}
}


#declare airlock = union {
  box {<-0.4, -0.1, -0.4> <0.4, 0.05, 0.4>}
  cylinder {0, 0.2*y, 0.3}
  torus {0.3, 0.025 translate 0.175*y}
  cylinder {-0.25*x, 0.25*x, 0.0125 rotate 20*y translate 0.2*y}
  cylinder {-0.25*x, 0.25*x, 0.0125 rotate 140*y translate 0.2*y}
  cylinder {-0.25*x, 0.25*x, 0.0125 rotate 260*y translate 0.2*y}
  texture {T_Hull}
}

#declare guns = union {

  #declare gun = difference {
    union {
      cylinder {0, 1*z, 0.1}
      cylinder {0.8*z, 1.01*z, 0.15}
    }
    cylinder {0.1*z, 1.1*z, 0.09}
    scale 0.6
  }

  difference {
    intersection {
      sphere {0 0.5 translate -0.2*y}
      box {<-0.5, -0.1, -0.5> <0.5, 1, 0.5>}
    }
    torus {0.4 0.025 translate 0.1*y texture {T_RedHull}}
  }
  box {<-0.2, 0, -0.35> <0.2, 0.25, -0.1>}
  object {gun translate 0.2*z rotate -30*x translate <-0.2, 0.1, 0>}
  object {gun translate 0.2*z rotate -30*x translate <0.2, 0.1, 0>}
  texture {T_Hull}
}

#declare antenna = union {
  difference {
    box {<-0.4, -0.1, -0.4> <0.4, 0.05, 0.4>}
    plane {-x-z, -0.3}
  }
  cone {0, 0.05, 0.6*y, 0}
  cylinder {0.5*y, 0.51*y, 0.15}
  cylinder {0.4*y, 0.41*y, 0.2}
  cylinder {0.3*y, 0.31*y, 0.25}
  texture {T_Hull}
}

#declare radar = union {
  
  #declare radar_scaffold = union {
    cylinder {-0.25*z, 0.4*y, 0.025}
    cylinder {0.25*z, 0.4*y, 0.025}
    sphere {0.4*y, 0.025}
  }

  #declare dish = union {
    difference {
      sphere {<0, 0, 2> 2}
      sphere {<0, 0, 2.5> 2.2}
    }
    cylinder {0, 2*z, 0.1}
    scale 0.15
  }

  cylinder {0, 0.05*y, 0.4}
  union {
    object {radar_scaffold translate -0.1*x}
    object {radar_scaffold translate 0.1*x}
    cylinder {<-0.1, 0.4, 0> <0.1, 0.4, 0> 0.025}
    object {dish rotate -20*x translate 0.4*y}
    translate 0.1*x
  }
  texture {T_Hull}
}

#declare tank = union {
  difference {
    box {<-0.4, -0.1, -0.4> <0.4, 0.05, 0.4>}
    plane {-x-z, -0.3}
  }
  difference {
    union {
      sphere {<-0.05, 0, -0.05> 0.3}
      sphere {<-0.05, 0.3, -0.05> 0.025}
    }
    plane {y 0}
  }
  texture {T_Hull}
}

#declare pipes = union {
  cylinder {<-0.3, -0.2, -0.4> <-0.3, 0.3, -0.4> 0.025}
  cylinder {<0.3, -0.2, -0.4> <0.3, 0.3, -0.4> 0.025}
  cylinder {<-0.3, 0.3, -0.4> <0.3, 0.3, -0.4> 0.025}
  sphere {<-0.3, 0.3, -0.4> 0.025}
  sphere {<0.3, 0.3, -0.4> 0.025}
  cylinder {<-0.3, 0, -0.1> <-0.3, 0, 0.3> 0.15}
  sphere {<-0.3, 0, -0.1> 0.15}
  sphere {<-0.3, 0, 0.3> 0.15}
  cylinder {<-0.0, 0, -0.1> <-0.0, 0, 0.3> 0.15}
  sphere {<-0.0, 0, -0.1> 0.15}
  sphere {<-0.0, 0, 0.3> 0.15}
  texture {T_Hull}
}


// Set up the array containing the techoration items
#declare items = array[10] {
  plate1,
  plate2,
  plate3,
  windows,
  airlock,
  guns,
  antenna,
  radar,
  tank,
  pipes
}
// Determine number of items; prevents us from having to manually
// change the following array declarations
#local nItems = dimension_size(items, 1);


// Do you wanna get an overview over the declared items?
#if (showItems)
  
  // Get union object containing all items nicely aligned in a square grid
  object {displayTechorationItems(items)
    rotate 30*x
    translate <0, 1, 0>
  }

  camera {
    location  <0, 0, -6>
    look_at   <0, 0, 0>
    right     x*image_width/image_height
  }

#else // No, thank you, just show me the techorated space ship

  // For each item, specify how to scale it
  #declare scaling = array[nItems] {
    <TECHORATE_SCALE_AVG, TECHORATE_SCALE_AVG, 0, 0>,
    <TECHORATE_SCALE_AVG, TECHORATE_SCALE_AVG, 0, 0>,
    <TECHORATE_SCALE_MIN, TECHORATE_SCALE_MIN, 0, 0>,
    <0, 0, TECHORATE_SCALE_MIN, 0>,
    <0, 0, 0, 0>,
    <0, 0, 0, 0>,
    <0, 0, TECHORATE_SCALE_MIN, 0>,
    <0, 0, TECHORATE_SCALE_AVG, TECHORATE_SCALE_AVG>,
    <0, 0, TECHORATE_SCALE_AVG, TECHORATE_SCALE_AVG>,
    <0, 0, TECHORATE_SCALE_AVG, TECHORATE_SCALE_AVG>
  }
  
  // For each item, specify how to rotate it, whether or not to cut it
  // to make it fit into its grid node and whether or not to cut it
  // out of (engrail into) the hull
  #declare modifications = array[nItems] {
    <0, 360, 4, TECHORATE_CUT_BOTH, 0>,
    <0, 360, 4, TECHORATE_CUT_BOTH, 0>,
    <0, 360, 4, 0, 0>,
    <0, 360, 4, 0, TECHORATE_ENGRAIL>,
    <0, 0, 0, 0, 0>,
    <-45, 45, 90, 0, 0>,
    <0, 360, 4, 0, 0>,
    <0, 360, 360, 0, 0>,
    <0, 360, 4, 0, 0>,
    <0, 360, 4, 0, 0>
  }
  
  // Pass items and what to do with them to the techorator
  setTechorationItems(items, scaling, modifications)
  
  // Specify how faces created by cutting / engrailing are to look like
  setTechorationCutoffTexture(T_Hull)

  // Set up a (vertical) cylindrical grid, scaled by a factor of 100 in x- and z direction
  setTechorationDimensions(TECHORATE_CYLINDRICAL_GEOMETRY, 0, 360, 18, -2, 6, 20, 100, 0, 100)

  // Tell the techorator that we want only "pipe" items for now
  #declare weights = array[nItems] {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 1
  };
  setTechorationWeights(weights)
  
  // Apply some double-sized pipes to the nozzle
  #declare nozzle = techorateInitialRandomGrid(object {nozzle rotate -90*z}, 0.1, 2, 2)

  // No we want both tanks and pipes
  #declare weights = array[nItems] {
    0, 0, 0, 0, 0, 0, 0, 0, 1, 1
  };
  setTechorationWeights(weights)
  
  // Add some normal-size techoration
  #declare nozzle = techorateNextRandomGrid(object {nozzle}, 0.5, 1, 1)
  
  // create an engine by combining nozzle, flame and some cool glow effects
  #declare engine = union {
    object {nozzle rotate 90*z}
    object{createGlow(0.5, 1.75, <0, 1, 1>, 0.5) rotate 90*y translate -2.75*x}
    object{createGlow(0.5, 0.9, <1, 1, 1>, 2) rotate 90*y translate -2.76*x}
    object {flame}
  }

  // Now for the actual vessel - set up another cylindrical grid
  setTechorationDimensions(TECHORATE_CYLINDRICAL_GEOMETRY, 0, 360, 36, -13, 13, 36, 400, 0, 1000)
  
  // Choose s random seed yielding some nice techoration
  setTechoratorRandomSeed(5);  // 5, 10, 44, 53, 80, 92

  // Initially, we only want some fancy red hull plates
  #declare weights = array[nItems] {
    1, 0, 0, 0, 0, 0, 0, 0, 0, 0
  };
  setTechorationWeights(weights)

  // Place some (very!) few quad-size hull plates
  #declare vessel4 = union {techorateInitialRandomGrid(vessel, 0.02, 4, 4)};
  
  // Now for some "normal" hull plates
  #declare weights = array[nItems] {
    0, 1, 0, 0, 0, 0, 0, 0, 0, 0
  };
  setTechorationWeights(weights)
  
  // Place some of them, double sized
  #declare vessel2 = union {techorateNextRandomGrid(vessel, 0.05, 2, 2)};
  
  // Now allow for (almost) everything to be placed
  #declare weights = array[nItems] {
    0, 0.05, 1, 0.25, 0.1, 0.1, 0.05, 0.1, 0.025, 0.025
  };
  setTechorationWeights(weights)

  // Fill up all unocciped grid positions with something
  #declare vessel = difference {
    union {
      object {vessel}
      object {vessel4}
      object {vessel2}
      union {finishGridTechoration(vessel)}
    }
    union {
      getTechorationCutoutItems()
    }
  }
  
  // Create the final vessel by adding two engines
  #declare vessel = union {
    object {vessel rotate -90*z }
    object {engine translate <-8.5, -0.5, -3>}
    object {engine translate <-8.5, -0.5, 3>}
  }
  
  // Include the finished vessel into the scene
  object {vessel}

  // Place the camera
  camera {
    location  camera_path(relMovieTime*20)
    look_at   <0, 0, 0>
    right     x*image_width/image_height
  }

#end

// Build a planet
sphere {<10000, -3000, 1000> 5000
  no_shadow
  texture {T_Planet}
}

// The planet's atmosphere
object {createGlow(5000, 6000, <0.9, 0.9, 1>, 0.5)
  rotate 90*x
  Point_At_Trans(<10000, -3000, 1000>)
  translate <10000, -3000, 1000>
}
                         