//---------------------------------------------------------------------
//
// "techorator.inc" -- Torsten Crass, 2008
//
//
//---------------------------------------------------------------------

#include "transforms.inc"

//=== Public constants ====================================

#declare TECHORATOR_VERSION = "1.1.0";

// Constants used for specifying which kind of grid is to be used
// for determining where to put techoration items
#declare TECHORATE_PLANAR_GEOMETRY = 0;
#declare TECHORATE_CYLINDRICAL_GEOMETRY = 1;
#declare TECHORATE_SPHERICAL_GEOMETRY = 2;

// Constants used for specifying whether a techoration item is to
// be scaled to the minimum, maximum or average size of its grid node
#declare TECHORATE_SCALE_MIN = 1;
#declare TECHORATE_SCALE_MAX = 2;
#declare TECHORATE_SCALE_AVG = 3;

// Constants specifying whether an item extending its grid node it
// to be cut to fit in A, B, or both grid directions
#declare TECHORATE_CUT_A = 1;
#declare TECHORATE_CUT_B = 1;
#declare TECHORATE_CUT_BOTH = 1;

// Constant specifying whether or not an item's shape will be cut
// out of (or engrailed into) the techorated surface, rather
// than being positioned on it
#declare TECHORATE_ENGRAIL = 1;

// Constant specifying that no maximum deviation between
// grid angle and surface normal has to be considered
#declare TECHORATE_NONE = -1;

//=== Internat globals ====================================

// Random seed / stream initially used for techorating
#declare _techoratorRandomSeed = 0;
#declare _techoratorRandomStream = seed(_techoratorRandomSeed);

// Vector taking up the techorated object's normal vector
// at the current grid position
#declare _techoratorCurrentNormal = <0, 0, 0>;
// Tells whether or not the current grid position is at all
// incident of the techorated objet's shape
#declare _techoratorCurrentHit = 0;

// Maximum deviation between grid angle and surface normal
#declare _techoratorMaxDeviation = TECHORATE_NONE;

//=== Public functions ====================================

// Specify techoration geometry
// 1.   geo:  techoration geometry
// 2.   A1:   start value for A direction
// 3.   A2:   end value for A direction
// 4.   nA:   number of steps in A direction
// 5.   B1:   start value for B direction
// 6.   B2:   end value for B direction
// 7.   nB:   number of steps in B direction
// 8.   sx:   scaling factor for techoration grid, x direction
// 9.   sy:   scaling factor for techoration grid, y direction
// 10.  sz:   scaling factor for techoration grid, z direction
#macro setTechorationDimensions(geo, A1, A2, nA, B1, B2, nB, sx, sy, sz)
  #declare _techoratorGeometry = geo;
  #declare _techorator_A1 = A1;
  #declare _techorator_A2 = A2;
  #declare _techorator_nA = nA;
  #declare _techorator_B1 = B1;
  #declare _techorator_B2 = B2;
  #declare _techorator_nB = nB;
  #declare _techorator_sx = sx;
  #declare _techorator_sy = sy;
  #declare _techorator_sz = sz;
#end

// Specify items to be used and how to apply them
//  1.  items:          array containing object declarations
//
//  2.  scaling:        array of 4-element vectors specifying how items
//                      may be scaled with respect to the item's grid node
//        x:            how to scale in A direction
//        y:            how to scale in B direction
//        z:            if != 0, min, max or average will be determined
//                      with regard to both A and B direction; x and
//                      y should then be set to 0
//        f:            if .z != 0, specifies how to scale in direction of
//                      surface normal
//  3.  modifications:  array containing 5-element vectors specifying further
//                      modifications
//        x:            minimum rotation angle
//        y:            maximum rotation angle
//        z:            number of rotation steps btw. min and max
//        f:            whether or not to cut item in A or B direction in
//                      order to fit them to a grid node
//        t:            if != 0, the objects will be cut out of rather than
//                      being sticked onto the surface
#macro setTechorationItems(items, scaling, modifications)
  #declare _techorationItems = items;
  #declare _techorationScaling = scaling;
  #declare _techorationModifications = modifications;
  #declare _techoration_nObj = dimension_size(_techorationItems, 1);
#end

// Specify relative frequencies for each techoration item
#macro setTechorationWeights(weights)
  #declare _techorationWeights = weights;
  #declare _techoration_cumWeights = array[_techoration_nObj];

  #local i = 0;
  #local s = 0;
  #while (i < _techoration_nObj)
    #local s = s + _techorationWeights[i];
    #declare _techoration_cumWeights[i] = s;
    #local i = i + 1;
  #end
  #local i = 0;
  #while (i< _techoration_nObj)
    #declare _techoration_cumWeights[i] = _techoration_cumWeights[i]/s;
    #local i = i + 1;
  #end
#end

// Specify textures for faces exposed by cutting items to make them
// fit into their grid node
#macro setTechorationCutoffTexture(cut)
  #declare _techorationCutoffTexture = cut;
#end

// Specify which random stream is to be used for techoration
#macro setTechoratorRandomStream(stream)
  #declare _techoratorRandomStream = stream;
#end

// Specify which random seed is to be used for techorator (creates
// new stream, too)
#macro setTechoratorRandomSeed(rseed)
  #declare _techoratorRandomSeed = rseed;
  #declare _techoratorRandomStream = seed(rseed);
#end

// Specify maximum angle between a grid point's direction vector
// and the corresponding surface normal. If this value is
// exceeded, the grid point will not be used for techoration.
#macro setTechoratorMaxDeviation(dev)
  #declare _techoratorMaxDeviation = dev;
#end

// Completely random techoration
//  1.  obj:    object to be techorated
//  2.  n:      number of items to be placed
#macro techorateRandomly(obj, n)
  _initRandomTechoration(n)
  _techorateRandomly(obj, n)
#end

// Techorate object by iterating over complete grid
#macro techorateCompleteGrid(obj)
  _initGridTechoration()
  _clearTechorationOccupancy()
  finishGridTechoration(obj)
#end

// Start randomly placing items on grid nodes
//  1.  obj:  object to be techorated
//  2.  tN:   techoration density; either number of items to be placed (tN >= 1)
//            or fraction of grid nodes to be occupied (tN < 1)
//  3.  ni:   specifies how many grid nodes every item will be scaled to extend over in A direction
//  4.  nj:   specifies how many grid nodes every item will be scaled to extend over in B direction
#macro techorateInitialRandomGrid(obj, tN, ni, nj)
  _initGridTechoration()
  _clearTechorationOccupancy()
  techorateNextRandomGrid(obj, tN, ni, nj)
#end

// Same as "techorateInitialRandomGrid", but used for further rounds
#macro techorateNextRandomGrid(obj, tN, ni, nj)
  _techorateRandomGrid(obj, tN, ni, nj)
#end

// Occupy all remaining grid nodes (if any) with items
#macro finishGridTechoration(obj)
  _techorateGrid(obj)
#end

// Returns  all those placed techoration items which habe
// been defined to be cut out of rather than placed onto
// the surface
#macro getTechorationCutoutItems()
  #local i = 0;
  #while (i < _techorate_nNodes)
    #ifdef (_techorate_itemTypes[i])
      #if (_techorationModifications[_techorate_itemTypes[i]].transmit)
        object {_techorate_items[i]}
      #end
    #end
    #local i = i + 1;
  #end
#end

// Returns an item to be used for next techoration; re-define when non-standard
// processing is required
//  1.  ttype:  item type (= index of item in item array)
//  2.  i:      grid coordinate, A direction component
//  3.  j:      grid coordinate, B direction component
//  4.  sVec:   where on the techorated object's surface the item will be placed
//  5.  nVec:   techorated object's normal vector at sVec
#macro getTechorationItem(ttype, i, j, sVec, nVec)
  #local tObj = getTechorationItem_techorator_native(ttype);
  tObj
#end

// Return a union containing all techoration items in a
// square grid arrangement (with some grid positions
// left blank if number of items isn't a square number)
#macro displayTechorationItems(items)
  #local n = dimension_size(items,1);
  #local s = 1.5;
  #local l = ceil(sqrt(n));

  union {
    #local i = 0;
    #while (i < l)
      #local j = 0;
      #while (j < l)
        #local k = l*i + j;
        #if (k < n)
          object {items[k] translate <j*s, 0, i*s>}
        #end
        #local j = j + 1;
      #end
      #local i = i + 1;
    #end
    translate <-(l-1)*s/2, 0, -(l-1)*s/2>
    rotate -90*x
  }
#end



//=== Internal functions ==================================


//--- Initialisation stuff

#macro _initRandomTechoration(d)
  #declare _techorate_nNodes = d;
  #declare _techorate_items = array[d];
  #declare _techorate_itemTypes = array[d];

  #declare _techorate_dA = (_techorator_A2 - _techorator_A1);
  #declare _techorate_dB = (_techorator_B2 - _techorator_B1);
#end

#macro _initGridTechoration()
  #declare _techorate_nNodes = _techorator_nA * _techorator_nB;
  #declare _techorate_items = array[_techorate_nNodes];
  #declare _techorate_itemTypes = array[_techorate_nNodes];

  #declare _techorate_dA = (_techorator_A2 - _techorator_A1) / _techorator_nA;
  #declare _techorate_dB = (_techorator_B2 - _techorator_B1) / _techorator_nB;

//  #declare _techoration_nObj = dimension_size(_techorationItems, 1);
#end


#macro _clearTechorationOccupancy()
  #declare _occupied = array[_techorate_nNodes];
  #local i = 0;
  #while (i < _techorate_nNodes)
    #declare _occupied[i] = 0;
    #local i = i + 1;
  #end
#end



//--- Functions iterating over grid

#macro _techorateRandomly(obj, d)
  #local k = 0;
  #while (k < d)
    #local rA = rand(_techoratorRandomStream);
    #local rB = rand(_techoratorRandomStream);
    #switch (_techoratorGeometry)
      #case (TECHORATE_SPHERICAL_GEOMETRY)
        #local A = _techorator_A1 + _techorate_dA*rA;
        #local y1 = sin(radians(_techorator_B1));
        #local y2 = sin(radians(_techorator_B2));
        #local r = y1 + (y2 - y1)*rB;
        #local B = degrees(asin(r));
      #break
      #case (TECHORATE_CYLINDRICAL_GEOMETRY)
        #local A = _techorator_A1 + _techorate_dA*rA;
        #local B = _techorator_B1 + _techorate_dB*rB;
      #break
      #case (TECHORATE_PLANAR_GEOMETRY)
        #local A = _techorator_A1 + _techorate_dA*rA;
        #local B = _techorator_B1 + _techorate_dB*rB;
      #break
    #end
    _selectNextRandomTechorationItem(A, B, k)
    #local k = k + 1;
  #end
  _applyTechorationToObject(obj, _techorate_nNodes)
#end

#macro _techorateGrid(obj)
  #local j = 0;
  #while (j < _techorator_nB)
    #local i = 0;
    #while (i < _techorator_nA)
       _selectNextGridTechorationItem(i, j, 1, 1)
       #local i = i + 1;
    #end
    #local j = j + 1;
  #end
  _applyTechorationToObject(obj, _techorate_nNodes)
#end


#macro _techorateRandomGrid(obj, tN, ni, nj)
  #switch (_techoratorGeometry)
    #case (TECHORATE_SPHERICAL_GEOMETRY)
      #local nA = _techorator_nA;
      #local nB = _techorator_nB;
    #break
    #case (TECHORATE_CYLINDRICAL_GEOMETRY)
      #local nA = _techorator_nA;
      #local nB = _techorator_nB + 1 - nj;
    #break
    #case (TECHORATE_PLANAR_GEOMETRY)
      #local nA = _techorator_nA + 1 - ni;
      #local nB = _techorator_nB + 1 - nj;
    #break
  #end

  #local nN = nA * nB;

  #local n = tN;
  #if (tN < 1)
    #local n = int(nN * tN);
  #end

  #local l = 0;
  #while (l < n)
    #local ok = 0;
    #while (!ok)
      #local i = int(rand(_techoratorRandomStream) * nA);
      #local j = int(rand(_techoratorRandomStream) * nB);
      #local ok = 1;
      #local li = 0;
      #while (li < ni)
        #local lj = 0;
        #while (lj < nj)
          #local lgi = _techorate_gridIndex(mod(i + li, _techorator_nA), mod(j + lj, _techorator_nB));
          #if (_occupied[lgi])
            #local ok = 0;
          #end
          #local lj = lj + 1;
        #end
        #local li = li + 1;
      #end
    #end
    _selectNextGridTechorationItem(i, j, ni, nj)
    #local l = l + 1;
  #end
  _applyTechorationToObject(obj, _techorate_nNodes)
#end


//--- Apply grid techoration to object

#macro _applyTechorationToObject(obj, n)
  #local i = 0;
  #while (i < n)
    #ifdef (_techorate_itemTypes[i])
      #if (!_techorationModifications[_techorate_itemTypes[i]].transmit)
        object {_techorate_items[i]}
      #end
    #end
    #local i = i + 1;
  #end
#end


//--- Getting techoration items

#macro getTechorationItem_techorator_native(ttype)
  #local tObj = _techorationItems[ttype];
  tObj
#end


#macro _selectNextRandomTechorationItem(A, B, k)
  #local gVec = _techorate_gridVec(A, B);
  #local bVec = _techorate_gridBaseVec(A, B);
  #local sVec = _techorate_gridSurfaceVec(bVec, gVec, obj);
  #if (_techoratorCurrentHit)
    #local nVec = _techoratorCurrentNormal;
    #local ttype = _techorate_randomIndex();

    #if (_techorationModifications[ttype].z != 0)
      #local rotation = _techorationModifications[ttype].x + int(_techorationModifications[ttype].z * rand(_techoratorRandomStream)) * (_techorationModifications[ttype].y - _techorationModifications[ttype].x)/_techorationModifications[ttype].z;
    #else
      #local rotation = 0;
    #end

    #local tObj = object {getTechorationItem(ttype, A, B, sVec, nVec) rotate y*rotation};
//    #declare tObj = object {tObj rotate -x * vang(y, nVec, 0) rotate -y * (90 + vang(x, <nVec.x, 0, nVec.z>, z)) translate sVec}
    #declare tObj = object {tObj
      Reorient_Trans(y, <0, nVec.y, -1>,)
      Reorient_Trans(-z, <nVec.x, 0, nVec.z>)
      translate sVec
    }
    #declare _techorate_items[k] = tObj;
    #declare _techorate_itemTypes[k] = ttype;
  #end
#end

#macro _selectNextGridTechorationItem(i, j, ni, nj)

  #local gi = _techorate_gridIndex(i, j);

  #if (!_occupied[gi])

    #local A = _techorator_A1 + _techorate_dA*i;
    #local B = _techorator_B1 + _techorate_dB*j;
    #local dA = _techorate_dA*ni;
    #local dB = _techorate_dB*nj;

    #local gVec = _techorate_gridVec(A+dA/2, B+dB/2);
    #local bVec = _techorate_gridBaseVec(A+dA/2, B+dB/2);
    #local sVec = _techorate_gridSurfaceVec(bVec, gVec, obj);

    #if (_techoratorCurrentHit)

      #local nVec = _techoratorCurrentNormal;

//printVecVar("angl", 180*acos(vdot(vnormalize(gVec), vnormalize(nVec)))/pi)

      #if ((_techoratorMaxDeviation < 0) | (180*acos(vdot(vnormalize(gVec), vnormalize(nVec)))/pi < _techoratorMaxDeviation))
        #local b00 = _techorate_gridBaseVec(A, B);
        #local g00 = _techorate_gridVec(A, B);
        #local v00 = _techorate_gridSurfaceVec(b00, g00, obj);
        #local h00 = _techoratorCurrentHit;

        #local b01 = _techorate_gridBaseVec(A, B+dB);
        #local g01 = _techorate_gridVec(A, B+dB);
        #local v01 = _techorate_gridSurfaceVec(b01, g01, obj);
        #local h01 = _techoratorCurrentHit;

        #local b10 = _techorate_gridBaseVec(A+dA, B);
        #local g10 = _techorate_gridVec(A+dA, B);
        #local v10 = _techorate_gridSurfaceVec(b10, g10, obj);
        #local h10 = _techoratorCurrentHit;

        #local b11 = _techorate_gridBaseVec(A+dA, B+dB);
        #local g11 = _techorate_gridVec(A+dA, B+dB);
        #local v11 = _techorate_gridSurfaceVec(b11, g11, obj);
        #local h11 = _techoratorCurrentHit;

        #if (h00 & h01 & h10 & h11)

          #local ttype = _techorate_randomIndex();
          #if (_techorationModifications[ttype].z != 0)
            #local rotation = _techorationModifications[ttype].x + int(_techorationModifications[ttype].z * rand(_techoratorRandomStream)) * (_techorationModifications[ttype].y - _techorationModifications[ttype].x)/_techorationModifications[ttype].z;
          #else
            #local rotation = 0;
          #end
          #local tObj = object {getTechorationItem(ttype, i, j, sVec, nVec) rotate y*rotation};


          #local wA1 = vlength(v10 - v00);
          #local wA2 = vlength(v11 - v01);
          #local wA = ni;
          #switch (_techorationScaling[ttype].x)
            #case (TECHORATE_SCALE_MIN)
              #local wA = min(wA1, wA2);
            #break
            #case (TECHORATE_SCALE_MAX)
              #local wA = max(wA1, wA2);
            #break
            #case (TECHORATE_SCALE_AVG)
              #local wA = (wA1 + wA2)/2;
            #break
          #end

          #local wB1 = vlength(v01 - v00);
          #local wB2 = vlength(v11 - v10);
          #local wB = nj;
          #switch (_techorationScaling[ttype].y)
            #case (TECHORATE_SCALE_MIN)
              #local wB = min(wB1, wB2);
            #break
            #case (TECHORATE_SCALE_MAX)
              #local wB = max(wB1, wB2);
            #break
            #case (TECHORATE_SCALE_AVG)
              #local wB = (wB1 + wB2)/2;
            #break
          #end

          #local sA = wA;
          #local sB = wB;
          #switch (_techorationScaling[ttype].z)
            #case (TECHORATE_SCALE_MIN)
              #local sA = sA * min(wA1, wA2, wB1, wB2);
              #local sB = sB * min(wA1, wA2, wB1, wB2);
            #break
            #case (TECHORATE_SCALE_MAX)
              #local sA = sA * max(wA1, wA2, wB1, wB2);
              #local sB = sB * max(wA1, wA2, wB1, wB2);
            #break
            #case (TECHORATE_SCALE_AVG)
              #local sA = sA * (wA1 + wA2 + wB1 + wB2)/4;
              #local sB = sB * (wA1 + wA2 + wB1 + wB2)/4;
            #break
          #end

          #local sC = 1;
          #if (_techorationScaling[ttype].z & _techorationScaling[ttype].filter)
            #switch (_techorationScaling[ttype].filter)
              #case (TECHORATE_SCALE_MIN)
                #local sC = sC * min(wA1, wA2, wB1, wB2);
              #break
              #case (TECHORATE_SCALE_MAX)
                #local sC = sC * max(wA1, wA2, wB1, wB2);
              #break
              #case (TECHORATE_SCALE_AVG)
                #local sC = sC * (wA1 + wA2 + wB1 + wB2)/4;
              #break
            #end
          #end

          #declare tObj = object {tObj
            scale <sA, sC, sB>
            Reorient_Trans(y, <0, nVec.y, -1>,)
            Reorient_Trans(-z, <nVec.x, 0, nVec.z>)
            translate sVec
          }

          #if (_techorationModifications[ttype].filter)

            #local bound = box {min_extent(tObj), max_extent(tObj)}

            #declare tObj = difference {
              object {tObj}
              union {
                #if ((_techorationModifications[ttype].filter = TECHORATE_CUT_A) | (_techorationModifications[ttype].filter = TECHORATE_CUT_BOTH))
                  #local d = b01 + g01 - (b00 + g00);
                  #if (vlength(d) != 0)
                    #local p = vnormalize(vcross(g00, d));
                    plane {-p, -vdot(b00, p) texture {_techorationCutoffTexture}}
                  #end
                  #local d = b10 + g10 - (b11 + g11);
                  #if (vlength(d) != 0)
                    #local p = vnormalize(vcross(g11, d));
                    plane {-p, -vdot(b11, p) texture {_techorationCutoffTexture}}
                  #end
                #end
                #if ((_techorationModifications[ttype].filter = TECHORATE_CUT_B) | (_techorationModifications[ttype].filter = TECHORATE_CUT_BOTH))
                  #local d = b10 + g10 - (b00 + g00);
                  #if (vlength(d) != 0)
                    #local p = vnormalize(vcross(g00, d));
                    plane {p, vdot(b00, p) texture {_techorationCutoffTexture}}
                  #end
                  #local d = b11 + g11 - (b01 + g01);
                  #if (vlength(d) != 0)
                    #local p = vnormalize(vcross(g11, d));
                    plane {-p, -vdot(b11, p) texture {_techorationCutoffTexture}}
                  #end
                #end
              }
              bounded_by {bound}
            }

          #end

          #declare _techorate_items[gi] = tObj;
          #declare _techorate_itemTypes[gi] = ttype;
          #local li = 0;
          #while (li < ni)
            #local lj = 0;
            #while (lj < nj)
              #local lgi = _techorate_gridIndex(mod(i + li, _techorator_nA), mod(j + lj, _techorator_nB));
              #declare _occupied[lgi] = 1;
              #local lj = lj + 1;
            #end
            #local li = li + 1;
          #end

        #end

      #end

    #end

  #end

#end


#macro _techorate_gridIndex(i, j)
  #local index =  i * _techorator_nB + j;
  index
#end


#macro _techorate_randomIndex()
  #local r = rand(_techoratorRandomStream);
  #local i = 0;
  #while (r > _techoration_cumWeights[i])
    #local i = i + 1;
  #end
  i
#end


//--- Vector maths


#macro _techorate_gridVec(A, B)

  #switch (_techoratorGeometry)
    #case (TECHORATE_PLANAR_GEOMETRY)
      #local vec = <0, 0, -_techorator_sz>;
    #break
    #case (TECHORATE_CYLINDRICAL_GEOMETRY)
      #local vec = <_techorator_sx*cos(radians(-A)), 0, _techorator_sz*sin(radians(-A))>;
    #break
    #case (TECHORATE_SPHERICAL_GEOMETRY)
      #local vec = <_techorator_sx*cos(radians(-A))*cos(radians(B)), _techorator_sy*sin(radians(B)), _techorator_sz*sin(radians(-A))*cos(radians(B))>;
    #break
  #end


  vec
#end


#macro _techorate_gridBaseVec(A, B)
  #switch (_techoratorGeometry)
    #case (TECHORATE_PLANAR_GEOMETRY)
      #local vec = <A, B, 0>;
    #break
    #case (TECHORATE_CYLINDRICAL_GEOMETRY)
      #local vec = <0, B, 0>;
    #break
    #case (TECHORATE_SPHERICAL_GEOMETRY)
      #local vec = <0, 0, 0>;
    #break
  #end
  vec
#end


#macro _techorate_gridSurfaceVec(center, vec, obj)
  #declare _techoratorCurrentHit = 1;
  #local sVec = trace(obj, center + vec, -vec, _techoratorCurrentNormal);
  #if (vlength(_techoratorCurrentNormal) = 0)
    #declare _techoratorCurrentHit = 0;
  #end
  #if (vdot(-vec, _techoratorCurrentNormal) > 0)
    #declare _techoratorCurrentNormal = -_techoratorCurrentNormal;
  #end
  sVec
#end
